#include <intercomm.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>


IC_Desc* create_ttable(int p, int r, float A[200]) {
  int globals[200], offsets[200], tasks[200];
  IC_Desc* desc;
  int i;
  
  /* assign this processor the rth portion of global index space */
  for (i = 0; i < 200; ++i) {
    globals[i] = 200*r + i;
  }
  
  /* round-robin global index to task assignment */
  for (i = 0; i < 200; ++i) {
    offsets[i] = i/p + 50*r;
    tasks[i] = i%p;
  }
  
  /* create the ic descriptor */
  desc = IC_Create_ttable_desc(globals, offsets, tasks, 200);
  
  /* assign each local value its global index */
  for (i = 0; i < 200; ++i) {
    A[i] = globals[i];
  }
  
  return desc;
}

/*
 * parameter
 *  argv[0] : program name
 *  argv[1] : local name
 *  argv[2] : other name
 *  argv[3] : local tasks
 *  argv[4] : other tasks
 *  argv[5] : machine file
 */
int main(int argc, char* argv[]) {
  char* localname; // = argv[1];
  char* othername; // = argv[2];
  int local_tasks; // = atoi(argv[3]);
  int other_tasks; // = atoi(argv[4]);
  int i, rank;

  MPI_Comm groupcomm;

  IC_Program* local;
  IC_Program* other;
  
  IC_Desc* desc;
  float A[200];
  
  IC_Region* region_set[2];
  int indices[8];
  IC_Sched* sched;
  int tag = 99;

  int numNode = -1;
  char** newargv;


  // initialize mpi
  printf("initialize mpi\n");
  MPI_Init(&argc, &argv);

  localname   = argv[1];
  othername   = argv[2];
  local_tasks = atoi(argv[3]);
  other_tasks = atoi(argv[4]);

  
  MPI_Comm_dup(MPI_COMM_WORLD, &groupcomm);
  MPI_Comm_rank(groupcomm, &rank);
  MPI_Barrier(groupcomm);

  
  
  /* initialize ic */
  printf("initialize ic\n");
  local = IC_Init(localname, local_tasks, rank);
  other = IC_Wait(othername, other_tasks);
  IC_Sync(local, other);

  desc = create_ttable(local_tasks, rank, A);
 
  /* define two 8 element regions for transfer */
  printf("define regions\n");
  for (i = 0; i < 8; ++i)
    indices[i] = i;
  region_set[0] = IC_Create_enum_region(indices, 8);
  for (i = 0; i < 8; ++i)
    indices[i] = i + 400;
  region_set[1] = IC_Create_enum_region(indices, 8);
  
  /* create a translation table type array descriptor and array */
  printf("create schedule\n");
  sched = IC_Compute_schedule(local, other, desc, region_set, 2);
  printf("send data\n");
  IC_Send_float(other, sched, &A[0], tag);
  IC_Sync(local, other);
  IC_Free_sched(sched);
  
  /* clean up */
  printf("clean up\n");
  IC_Free_region(region_set[1]);
  IC_Free_region(region_set[0]);

  IC_Free_desc(desc);

  IC_Free_program(other);
  IC_Quit(local);

  MPI_Barrier(groupcomm);
  MPI_Comm_free(&groupcomm);
  MPI_Finalize();
  
  return 0;
}
