/**
 * $Author: iyoon $
 * $Date: 2009/10/01 21:56:18 $
 * $Revision: 1.2 $
 * $Source: /fs/cvs/chaos/InterComm/examples/rainbow/black.c,v $
 * 
 * Written by Norman Lo <normanlo@cs.umd.edu>
 */

#include <intercomm.h>
#include <mpi.h>
#include <stdio.h>
#include <unistd.h>

#define COMPONENT_NAME "blackcomp"
#define REGION_NAME1   "blackR1"


/**
 * Local view
 * Initialize 1-dim float array
 */

void print_array(int rank, float data[2][2]){
    int i, j;
    
    fprintf(stdout, "My rank is : %d\n", rank);
    for( i = 0 ; i < 2 ; i++ ){
        for( j = 0 ; j < 2 ; j++){
            fprintf(stdout, "%4.0f  ", data[i][j]);
        }
        fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n");
}

int parse_xjdname(int argc, char** argv, char** xjdname){
    int c; 
    int xjdflag = 0;
    
    optind = 1; /* reset option index to 1 to parse from the first option*/
 
    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'd':
                *xjdname = (char*)strdup(optarg);
    
                xjdflag = 1;
                break;
        }
    }
    
    if(!xjdflag){
        return -1;
    }
    else{
        return 0;
    }
}   
    





/**
 * Global view
 * This part should be decided by application programmer
 */

IC_Desc* create_ttable(int ntask, int rank, float A[2][2]){
    int i;
    int globals[4], offsets[4], tasks[4];
    IC_Desc* desc;

    for (i=0; i<4; i++){
        globals[i]=4*rank+i;
        offsets[i] = i;
        tasks[i] = (4*rank+i)/ntask;
    }

    fprintf(stderr,"myrank = %d\n", rank);
    fprintf(stderr,"------------------------------\n");

    for (i=0; i<4; i++){
        fprintf(stderr,"i = %d, globals = %d, offsets = %d, tasks = %d\n", i, globals[i], offsets[i], tasks[i]);
   }

    desc = IC_Create_ttable_desc(globals, offsets, tasks, 4);

    return desc;
}

/**
 * This part should be decided by application programmer
 */

void initMPI(int* argc, char*** argv, int* rank, MPI_Comm* groupcomm)
{
    MPI_Init(argc, argv);
    MPI_Comm_dup(MPI_COMM_WORLD, groupcomm);
    MPI_Comm_rank(*groupcomm, rank);
    MPI_Barrier(*groupcomm);
}

void finishMPI(MPI_Comm* groupcomm)
{
    MPI_Barrier(*groupcomm);
    MPI_Comm_free(groupcomm);
    MPI_Finalize();
}

/*
 * execution
 * mpirun -np 4 ./black -d full_path_name_of_xjd.xml 
 * 
 * this component expose...
 *  name: blackcomp
 *  dataport: blackR1
 */
int main(int argc, char **argv)
{
    IC_XJD* xjd;
    
    float A[2][2];
    IC_Desc *desc;
    IC_Region* region1[2];
    int indices1[4], indices2[2];

    MPI_Comm groupcomm;
    int rank, status;
    char* xjdname;

    int i;

    /* initialize mpi and data*/
    fprintf(stderr, "initialize mpi and getting rank\n");
    initMPI(&argc, &argv, &rank, &groupcomm);

    fprintf(stderr, "initialize InterComm\n");    

    /* XJD name assignment should be done after MPI_Init*/
    if(parse_xjdname(argc, argv, &xjdname) < 0){
        fprintf(stderr, "XJD Option is required...\n");
        return -1;
    }

    if ( (xjd = IC_Initialize(COMPONENT_NAME, rank, xjdname, &status)) == NULL ){
        fprintf(stderr, "IC_Initialize failed at rank : %d\n", rank);        
        return -1;
    }


    desc = create_ttable (4, rank, A);

    for (i=0; i<4; i++)
       indices1[i] = i;
    for (i=0; i<2; i++)
       indices2[i] = i+5;

    region1[0] = IC_Create_enum_region(indices1, 4);
    region1[1] = IC_Create_enum_region(indices2, 2);

    IC_Register_region(xjd, region1, 2, REGION_NAME1, desc, A, &status);

    IC_Commit_region(xjd, &status);
    fprintf (stderr, "After commiting, Status = %d\n", status);

    status = IC_Import(xjd, REGION_NAME1);

    print_array(rank, A);

    /* free allocated memory*/
    fprintf(stderr, "clean up\n");
    IC_Free_region(region1[0]);
    IC_Free_region(region1[1]);
    IC_Free_desc(desc); 

    IC_Finalize(xjd, &status);

    fprintf(stderr, "finalize mpi\n");
    finishMPI(&groupcomm);

    return 0;
}
