/**
 * $Author: iyoon $
 * $Date: 2009/10/01 21:56:18 $
 * $Revision: 1.2 $
 * $Source: /fs/cvs/chaos/InterComm/examples/rainbow/green.c,v $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */

#include <intercomm.h>
#include <stdio.h>
#include <unistd.h>
#include "pvm_tls.h"

#define PVM_GROUP      "GreenPvm"
#define COMPONENT_NAME "greencomp"
#define REGION_NAME1   "greenR1"

/**
 * Local view
 * Initialize 1-dim float array
 */
void init_data(int rank, float data[2]) {
    int i ;

    fprintf(stderr, "initialize data\n");
    for (i = 0 ;  i < 2 ; i++){
        data[i] = -(rank*100 + i + 1);
    }
}

void print_array(int rank, float data[2]){
    int i;

    fprintf(stdout, "My rank is : %d\n", rank);
    for( i = 0 ; i < 2 ; i++ ){
        fprintf(stdout, "%4.0f  ", data[i]);
    }
    fprintf(stdout, "\n");
}

    
int parse_xjdname(int argc, char** argv, char** xjdname){
    int c; 
    int xjdflag = 0;
    
    optind = 1; /* reset option index to 1 to parse from the first option*/
    
    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'd':
                *xjdname = (char*)strdup(optarg);

                xjdflag = 1;
                break;
        }
    }
    
    if(!xjdflag){
        return -1;
    }
    else{
        return 0;
    }
}   
    


/**
 * Global view
 * This part should be decided by application programmer
 */
IC_Desc *create_block() {
    int blocks[2][2][1] = {
        { {0}, {1} },
        { {2}, {3} }
    };
    
    int tasks[2]= {0, 1};
    int stride[2]={1, 1};
    
    return IC_Create_bdecomp_desc(1, &blocks[0][0][0], tasks, 2, IC_ROW_MAJOR);
}

/**
 * This part should be decided by application programmer
 */
IC_Region* create_region1(){
    int i;
    int lower[1], upper[1], stride[1];

    for (i = 0; i < 1; ++i) {
        lower[i] = 0;
        upper[i] = 3;
        stride[i] = 1;
    }
    
    return IC_Create_block_region(1, lower, upper, stride);  
}

void finishPVM(char* grpname, int my_nTask){
    pvm_barrier(grpname, my_nTask);
    pvm_lvgroup(grpname);
    pvm_exit();
}

/*
 * execution
 * $PVM_BIN/green -n 2 -d full_path_name_of_xjd.xml 
 * 
 * this component expose...
 * 	name: greencomp
 * 	dataport: greenR1
 */
int main(int argc, char **argv) {
    
    IC_XJD* xjd;
    
    float local_data[2];
    IC_Desc *desc;
    IC_Region* region1[1];

    int rank, my_nTask, status;
    char* xjdname;

    if(parse_xjdname(argc, argv, &xjdname) < 0){
        fprintf(stderr, "XJD Option is required...\n");
        return -1;
    }

    fprintf(stderr, "initialize pvm and getting rank\n");
    if( (init_pvm_mfile(argc, argv, PVM_GROUP, &my_nTask, &rank) ) < 0){
      fprintf(stderr, "Error in initializing PVM...\n");
      return -1;
    }

    fprintf(stderr, "initialize InterComm\n");   
    if ( (xjd = IC_Initialize(COMPONENT_NAME, rank, xjdname, &status)) == NULL ){
        fprintf(stderr, "IC_Initialize failed at rank : %d\n", rank);        
        return -1;
    }

    /* Initialize local block and set global view*/
    init_data(rank, local_data);
    desc = create_block();

    /* create and register region*/
    region1[0] = create_region1();
    IC_Register_region(xjd, region1, 1, REGION_NAME1, desc, local_data, &status);

    /* all regions are registered. Now.. we can export or import    */
    IC_Commit_region(xjd, &status);
 
    /* export region.*/
    status = IC_Export(xjd, REGION_NAME1);  

    /* print result array */
    print_array(rank, local_data);
 
    /* free allocated memory*/
    fprintf(stderr, "clean up\n");
    IC_Free_region(region1[0]);
    IC_Free_desc(desc); 
    IC_Finalize(xjd, &status);

    finishPVM(PVM_GROUP, my_nTask);

    return 0;
}
