/**
 * $Author: iyoon $
 * $Date: 2009/10/01 21:56:18 $
 * $Revision: 1.2 $
 * $Source: /fs/cvs/chaos/InterComm/examples/rainbow/red.c,v $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */

#include <intercomm.h>
#include <mpi.h>
#include <stdio.h>
#include <unistd.h>

#define COMPONENT_NAME "redcomp"
#define REGION_NAME1   "redR1"


/**
 * Local view
 * Initialize 1-dim float array
 */
void init_data(int rank, float data[2][2]) {
    int i,j;
    
    fprintf(stderr, "initialize data\n");
    for (i = 0 ;  i < 2 ; i++){
        for (j = 0 ; j < 2 ; j++){
            data[i][j] = -(rank*1000 + i*100 + j*10);
        }
    }
}

void print_array(int rank, float data[2][2]){
    int i, j;
    
    fprintf(stdout, "My rank is : %d\n", rank);
    for( i = 0 ; i < 2 ; i++ ){
        for( j = 0 ; j < 2 ; j++){
            fprintf(stdout, "%4.0f  ", data[i][j]);
        }
        fprintf(stdout, "\n");
    }
    fprintf(stdout, "\n");
}

int parse_xjdname(int argc, char** argv, char** xjdname){
    int c; 
    int xjdflag = 0;
    
    optind = 1; /* reset option index to 1 to parse from the first option*/
 
    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'd':
                *xjdname = (char*)strdup(optarg);
    
                xjdflag = 1;
                break;
        }
    }
    
    if(!xjdflag){
        return -1;
    }
    else{
        return 0;
    }
}   
    

/**
 * Global view
 * This part should be decided by application programmer
 */
IC_Desc *create_block() {
    int blocks[4][2][2] = {
        { {0, 0}, {1, 1} },
        { {2, 0}, {3, 1} },
        { {4, 0}, {5, 1} },
        { {6, 0}, {7, 1} },
    };
    
    int tasks[4]= {0, 1, 2, 3};
    int stride[4] = {1,1,1,1};
    
    /*return IC_Create_bdecomp_desc(2, &blocks[0][0][0], tasks, 4, stride, IC_ROW_MAJOR);*/
    return IC_Create_bdecomp_desc(2, &blocks[0][0][0], tasks, 4, IC_ROW_MAJOR);
}

/**
 * This part should be decided by application programmer
 */
void create_region1(IC_Region* region[2]){
    int lower[2], upper[2], stride[2];

    lower[0] = 0;
    lower[1] = 0;
    upper[0] = 1;
    upper[1] = 1;
    stride[0] = 1;
    stride[1] = 1;
    region[0] = IC_Create_block_region(2, lower, upper, stride);
    
    lower[0] = 6;
    lower[1] = 0;
    upper[0] = 7;
    upper[1] = 0;
    stride[0] = 1;
    stride[1] = 1;
    region[1] = IC_Create_block_region(2, lower, upper, stride);
}

void initMPI(int* argc, char*** argv, int* rank, MPI_Comm* groupcomm)
{
    MPI_Init(argc, argv);
    MPI_Comm_dup(MPI_COMM_WORLD, groupcomm);
    MPI_Comm_rank(*groupcomm, rank);
    MPI_Barrier(*groupcomm);
}

void finishMPI(MPI_Comm* groupcomm)
{
    MPI_Barrier(*groupcomm);
    MPI_Comm_free(groupcomm);
    MPI_Finalize();
}

/*
 * execution
 * mpirun -np 4 ./red -d full_path_name_of_xjd.xml 
 * 
 * this component expose...
 *  name: redcomp
 *  dataport: redR1
 */
int main(int argc, char **argv)
{
    IC_XJD* xjd;
    
    float local_data[2][2];
    IC_Desc *desc;
    IC_Region* region1[2];

    MPI_Comm groupcomm;
    int rank, status;
    char* xjdname;

    /* initialize mpi and data*/
    fprintf(stderr, "initialize mpi and getting rank\n");
    initMPI(&argc, &argv, &rank, &groupcomm);

    fprintf(stderr, "initialize InterComm\n");    

    /* XJD name assignment should be done after MPI_Init*/
    if(parse_xjdname(argc, argv, &xjdname) < 0){
        fprintf(stderr, "XJD Option is required...\n");
        return -1;
    }

    if ( (xjd = IC_Initialize(COMPONENT_NAME, rank, xjdname, &status)) == NULL ){
        fprintf(stderr, "IC_Initialize failed at rank : %d\n", rank);        
        return -1;
    }

    /* Initialize local block and set global view*/
    init_data(rank, local_data);
    desc = create_block();

    /* create and register region*/
    create_region1(region1);
    IC_Register_region(xjd, region1, 2, REGION_NAME1, desc, local_data, &status);
   
    /* all regions are registered. Now.. we can export or import    */
    IC_Commit_region(xjd, &status);

    /* export region.*/
    status = IC_Export(xjd, REGION_NAME1);

    /* print result array */
    print_array(rank, local_data);

    /* free allocated memory*/
    fprintf(stderr, "clean up\n");
    IC_Free_region(region1[0]);
    IC_Free_region(region1[1]);
    IC_Free_desc(desc); 

    IC_Finalize(xjd, &status);

    fprintf(stderr, "finalize mpi\n");
    finishMPI(&groupcomm);

    return 0;
}
