divert(-1)
/*//// M4 macros for P++ methods*/

define(`PPP_EXPORT_HELPER',`
void IC_EndPoint::exportArray(const $1Array& a, int msgtag, int& status) {
/* exportArrays a double array using InterComm*/
  unsigned ndim=a.numberOfDimensions();
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_INVALID_NDIM] << endl;
#endif
    status=IC_INVALID_NDIM;
    return;
  }
  unsigned i;
  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];
  
  for(i=0;i<ndim;++i) {
    lower[i]=a.getBase(i); 
    upper[i]=a.getBound(i); 
    stride[i]=a.getStride(i);

#ifdef IC_DEBUG
    cout << "P++ArrayInfo : " << lower[i] << " " << upper[i] << " " << stride[i] << endl;
#endif
  }

#ifdef IC_DEBUG
  cout << "--- number of dimensions " << ndim << endl;
#endif


  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=getPartiDescriptor(a); /*,thisProgramAO==otherProgramAO);*/
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Send_$1(otherProgram,sched,a.getLocalArray().getDataPointer(), msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;
}')

define(`PPP_IMPORT_HELPER',`
void IC_EndPoint::importArray(const $1Array& a, int msgtag, int& status) {
/* importArrays a double array using InterComm*/
  unsigned ndim=a.numberOfDimensions();
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_INVALID_NDIM] << endl;
#endif
    status=IC_INVALID_NDIM;
    return;
  }
  unsigned i;
  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];

  for(i=0;i<ndim;++i) {
    lower[i]=a.getBase(i); 
    upper[i]=a.getBound(i); 
    stride[i]=a.getStride(i);

#ifdef IC_DEBUG
    cout << "P++ArrayInfo : " << lower[i] << " " << upper[i] << " " << stride[i] << endl;
#endif
  }

#ifdef IC_DEBUG
  cout << "--- number of dimensions is " << ndim << endl;
#endif

  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=getPartiDescriptor(a); /*,thisProgramAO==otherProgramAO);*/
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Recv_$1(otherProgram,sched,a.getLocalArray().getDataPointer(), msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;
}'
)

define(`PPP_BCAST_LOCAL_HELPER',`
void IC_EndPoint::bcastLocalArray(const $1Array& a, int nelems, int tag, int& status) {
  
  status = IC_Bcast_local_$1(otherProgram, a.getLocalArray().getDataPointer(), nelems, tag); 

  if(status != 1){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)

define(`PPP_RECV_LOCAL_HELPER',`
void IC_EndPoint::recvLocalArray(const $1Array& a, int nelems, int tag, int& status) {
 
  status = IC_Recv_local_$1(otherProgram, a.getLocalArray().getDataPointer(), nelems, tag); 

  if(status != 1){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)


define(`PPP_ICXJD_BCAST_LOCAL_HELPER',`
void IC_EndPointSet::bcastLocalArray(const char* ptname, const $1Array& arr, int nelems, int& stat){ 
#ifdef IC_DEBUG
  cout << "Bcast Port" << endl;
#endif
  map<string, IC_Connection*>::const_iterator citer;
  for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
    string connname = citer->first;    
    IC_Connection* conn = (IC_Connection*)citer->second;
      
    if(!strcmp(conn->get_ctype().c_str(), "MxN")){ 
#ifdef IC_DEBUG
      cout << "\tskip MxN connection: " << connname << endl;
#endif
      continue;
    }

    if(conn->get_portname().compare(ptname) != 0){ // different portname
#ifdef IC_DEBUG
      cout << "\tskip different port name..." << endl;
#endif
      continue;
    }

    if(conn->get_type() != 0){ // connection is not for bcast
#ifdef IC_DEBUG
      cout << "\tskip port for recv data" << endl;
#endif
      continue;
    }
        
    IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
    IC_Program* myprog = ep->get_myprog();
    IC_Program* otherprog = ep->get_otherprog();
    IC_Sched* sched = conn->get_sched();
    int tag = conn->get_tag();
     
    cout << "\tOk... Broadcast : " << ptname 
         << " to " << otherprog->name << endl;

    stat = IC_Bcast_local_$1(otherprog, arr.getLocalArray().getDataPointer(), nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }
  } 

  stat = IC_OK;
}'
)

define(`PPP_ICXJD_RECV_LOCAL_HELPER',`
void IC_EndPointSet::recvLocalArray(const char* ptname, const $1Array& arr, int nelems, int& stat){ 
  cout << "Recv (bcasted) port" << endl;

  map<string, IC_Connection*>::const_iterator citer;
  for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
    string connname = citer->first;    
    IC_Connection* conn = (IC_Connection*)citer->second;
      
    if(!strcmp(conn->get_ctype().c_str(), "MxN")){ 
#ifdef IC_DEBUG
      cout << "\tskip MxN connection: " << connname << endl;
#endif
      continue;
    }

    if(conn->get_portname().compare(ptname) != 0){ // different portname
#ifdef IC_DEBUG
      cout << "\tskip different port name..." << endl;
#endif
      continue;
    }

    if(conn->get_type() != 1){ // connection is not for import
#ifdef IC_DEBUG
      cout << "\tskip port for bcast data" << endl;
#endif
      continue;
    }
        
    IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
    IC_Program* myprog = ep->get_myprog();
    IC_Program* otherprog = ep->get_otherprog();
    IC_Sched* sched = conn->get_sched();
    int tag = conn->get_tag();
     
    cout << "\tOk... Recv bcasted : " << ptname 
         << " from " << otherprog->name << endl;

    stat = IC_Recv_local_$1(otherprog, arr.getLocalArray().getDataPointer(), nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    }
  }

  stat = IC_OK;
}'
)


define(`PPP_ARRAY_REGISTER_HELPER',`
void IC_EndPointSet::registerArray(const char* ptname, const $1Array& arr, int& stat){     
    #ifdef IC_DEBUG
      cout << "Registering Port " << ptname << " with actual array." << endl;
    #endif

    string portname = ptname;
    stat = IC_ARRAY_REGISTER_FAILURE;        

    map<string, const $1Array*>::const_iterator iter;
    iter = this->ppp_$1_arrmap.find(portname);

    if(iter == this->ppp_$1_arrmap.end()){
        #ifdef IC_DEBUG
            cout << "\tno such key in ppp_$1_arrmap..." << endl;
        #endif

        stat = IC_ARRAY_REGISTER_FAILURE;
        return;
    }
    else{
        /* bind given array into the array map        */
        #ifdef IC_DEBUG
            cout << "\tportname found in ppp_$1_arrmap..." << endl;
        #endif

        this->ppp_$1_arrmap[portname] = &arr;
        stat = IC_OK;
    }

    if(stat == IC_OK){
        unsigned ndim=arr.numberOfDimensions();

        if (ndim > IC_MAX_DIM) {
            #ifdef IC_DEBUG
                cout << "\t" << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
            #endif
            stat = IC_EndPoint::IC_INVALID_NDIM;
            return;
        }

        unsigned i;
        int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];
  
        for(i=0;i<ndim;++i) {
            lower[i]=arr.getBase(i); 
            upper[i]=arr.getBound(i); 
            stride[i]=arr.getStride(i);
            
            #ifdef IC_DEBUG
                cout << "\t" << lower[i] << " " << upper[i] << " " << stride[i] << endl;
            #endif
        }

        #ifdef IC_DEBUG
            cout << "\tnumber of dimensions " << ndim << endl;
        #endif

        map<string, IC_Connection*>::const_iterator citer;
        for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
            string connname = citer->first;    
  
            IC_Connection* conn = (IC_Connection*)citer->second;
#ifdef IC_DEBUG
            cout << "\tCurrent..." << connname << " : "
                 << conn->get_portname() << endl;
#endif              
            
            if(!strcmp(conn->get_ctype().c_str(), "1xN")){
#ifdef IC_DEBUG
              cout << "\tSkipping 1xN connection: " << connname << endl;
#endif              
              continue;
            }

            if(conn->get_portname().compare(ptname) != 0){
#ifdef IC_DEBUG
              cout << "\tSkipping port" << endl;
#endif              
              continue;
            }

            IC_Region** region_set = (IC_Region**)malloc(sizeof(IC_Region*));
            region_set[0] = IC_Create_block_region(ndim, lower, upper, stride);
            if (!region_set[0]) {
              #ifdef IC_DEBUG
                cout << "\t" << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
              #endif
            
              stat = IC_EndPoint::IC_CANT_ALLOC_REGION;
              return;
            }
            else{
        	    conn->set_rgn(region_set);
            }

            IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
            unsigned myao = ep->get_myprog_arrayorder();
            unsigned oao  = ep->get_otherprog_arrayorder();
            #ifdef IC_DEBUG
                cout << "\tArray Order : " << myao << " " << oao <<endl;
            #endif
            
            IC_Desc* desc = getPartiDescriptor(arr); /*, myao==oao);*/      

            if (!desc) {
                #ifdef IC_DEBUG
                    cout << "\t" << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
                #endif
                stat = IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
                return;
            }
            else{
                conn->set_desc(desc);
            }
        }
    }
    else{
        stat = IC_ARRAY_REGISTER_FAILURE;        
    }
}'
)


/*//// M4 macros for Fortran 90 methods*/



define(`F90_EXPORT_HELPER',`
void IC_EndPoint::exportArray(const F90_$1Array& a, int msgtag, int& status) {

/* exportArrays a double array using InterComm*/
  unsigned i;
  unsigned ndim=a.rank;
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
#endif
    status=IC_EndPoint::IC_INVALID_NDIM;
    return;
  }

  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];

  for(i=0;i<ndim;++i) {
    lower[i]=0;
    upper[i]= a.shape[i] - 1;
    stride[i]= 1;

#ifdef IC_DEBUG    
    printf("Export: F90Array Structure: shape [%d] %d stride [%d] %d\n",i,a.shape[i],i,a.stride[i]);
#endif
  }

#ifdef IC_DEBUG    
  cout << "number of dimensions is " << ndim << endl;
#endif


  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_EndPoint::IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=buildDescriptor(a);
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Send_$1(otherProgram,sched,a.base, msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
    status=IC_EndPoint::IC_COMM_FAILURE;
    return;
  }

  IC_Free_sched(sched);
  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;

}')

define(`F90_IMPORT_HELPER',`
void IC_EndPoint::importArray(const F90_$1Array& a, int msgtag, int& status) {

/* importArrays a double array using InterComm*/
  unsigned ndim=a.rank;
  if (ndim>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
#endif
    status=IC_EndPoint::IC_INVALID_NDIM;
    return;
  }
  unsigned i;
  int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];

  for(i=0;i<ndim;++i) {
    lower[i]=0;
    upper[i]= a.shape[i] - 1;
    stride[i]= 1;

#ifdef IC_DEBUG
    printf("Import: F90Array Structure: shape [%d] %d stride [%d] %d\n",i, a.shape[i],i,a.stride[i]);
#endif
  }

#ifdef IC_DEBUG
  cout << "number of dimensions is " << ndim << endl;
#endif

  IC_Region* region_set[1];
  region_set[0]=IC_Create_block_region(ndim,lower,upper,stride);
  if (!region_set[0]) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
#endif
    status=IC_EndPoint::IC_CANT_ALLOC_REGION;
    return;
  }

  IC_Desc* desc=buildDescriptor(a);
  if (!desc) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
#endif
    status=IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
    return;
  }

  /* this can probably be cached in a more sophisticated version!!!*/
  IC_Sched* sched=IC_Compute_schedule(thisProgram,otherProgram,desc,region_set,1);
  if (!sched) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE] << endl;
#endif
    status=IC_EndPoint::IC_CANT_COMPUTE_COMM_SCHEDULE;
    return;
  }

  int ret=IC_Recv_$1(otherProgram,sched,a.base, msgtag);
  if (ret!=0) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
#endif
    status=IC_EndPoint::IC_COMM_FAILURE;
    return;
  }
 
  IC_Free_sched(sched);
  IC_Free_region(region_set[0]);
  IC_Free_desc(desc);

  status=IC_OK;

}')


define(`F90_BCAST_LOCAL_HELPER',`
void IC_EndPoint::bcastLocalArray(const F90_$1Array& a, int nelems, int tag, int& status) {
  
  status = IC_Bcast_local_$1(otherProgram, a.base, nelems, tag); 

  if(status != 0){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)

define(`F90_RECV_LOCAL_HELPER',`
void IC_EndPoint::recvLocalArray(const F90_$1Array& a, int nelems, int tag, int& status) {
  
  status = IC_Recv_local_$1(otherProgram, a.base, nelems, tag); 

  if(status != 0){
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_COMM_FAILURE] << endl;
#endif
    status=IC_COMM_FAILURE;
    return;
  }

  status = IC_OK;
}'
)


define(`F90_ICXJD_BCAST_LOCAL_HELPER',`
void IC_EndPointSet::bcastLocalArray(const char* ptname, const F90_$1Array& arr, int nelems, int& stat){ 
#ifdef IC_DEBUG
  cout << "Bcast Port" << endl;
#endif

  map<string, IC_Connection*>::const_iterator citer;
  for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
    string connname = citer->first;    
    IC_Connection* conn = (IC_Connection*)citer->second;
      
    if(!strcmp(conn->get_ctype().c_str(), "MxN")){ 
#ifdef IC_DEBUG
      cout << "\tskip MxN connection: " << connname << endl;
#endif
      continue;
    }

    if(conn->get_portname().compare(ptname) != 0){ // different portname
#ifdef IC_DEBUG
      cout << "\tskip different port name..." << endl;
#endif
      continue;
    }

    if(conn->get_type() != 0){ // connection is not for bcast
#ifdef IC_DEBUG
      cout << "\tskip port for recv data" << endl;
#endif
      continue;
    }
        
    IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
    IC_Program* myprog = ep->get_myprog();
    IC_Program* otherprog = ep->get_otherprog();
    IC_Sched* sched = conn->get_sched();
    int tag = conn->get_tag();
     
    cout << "\tOk... Broadcast : " << ptname 
         << " to " << otherprog->name << endl;

    stat = IC_Bcast_local_$1(otherprog, arr.base, nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    } 
  }

  stat = IC_OK;
}'
)

define(`F90_ICXJD_RECV_LOCAL_HELPER',`
void IC_EndPointSet::recvLocalArray(const char* ptname, const F90_$1Array& arr, int nelems, int& stat){ 
#ifdef IC_DEBUG
  cout << "Recv (bcasted) port" << endl;
#endif 

  map<string, IC_Connection*>::const_iterator citer;
  for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
    string connname = citer->first;    
    IC_Connection* conn = (IC_Connection*)citer->second;
      
    if(!strcmp(conn->get_ctype().c_str(), "MxN")){ 
#ifdef IC_DEBUG
      cout << "\tskip MxN connection: " << connname << endl;
#endif
      continue;
    }

    if(conn->get_portname().compare(ptname) != 0){ // different portname
#ifdef IC_DEBUG
      cout << "\tskip different port name..." << endl;
#endif
      continue;
    }

    if(conn->get_type() != 1){ // connection is not for import
#ifdef IC_DEBUG
      cout << "\tskip port for bcast data" << endl;
#endif
      continue;
    }
        
    IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
    IC_Program* myprog = ep->get_myprog();
    IC_Program* otherprog = ep->get_otherprog();
    IC_Sched* sched = conn->get_sched();
    int tag = conn->get_tag();
     
    cout << "\tOk... Recv bcasted : " << ptname 
         << " from " << otherprog->name << endl;

    stat = IC_Recv_local_$1(otherprog, arr.base, nelems, tag);
    if (stat != 1) {
        #ifdef IC_DEBUG
            cout << IC_EndPoint::IC_errors[-IC_EndPoint::IC_COMM_FAILURE] << endl;
        #endif
        stat = IC_EndPoint::IC_COMM_FAILURE;
        return;
    } 
  }

  stat = IC_OK;
}'
)

define(`F90_BUILD_DESCRIPTOR_HELPER',`
IC_Desc* IC_EndPoint::buildDescriptor(const F90_$1Array& a) {
  if (static_cast<unsigned>(a.rank)>IC_MAX_DIM) {
#ifdef IC_DEBUG
    cout << IC_EndPoint::IC_errors[-IC_INVALID_NDIM] << endl;
#endif
    return NULL;
  }

  int tasks[1];
  tasks[0]=0;

  IC_Desc* desc;

  switch(a.rank) {
    case 1: {
      int blocks[1][2][1];
      int blockstride[1];

      blocks[0][0][0]=0;
      blocks[0][1][0]=a.shape[0]-1;

      blockstride[0] = a.stride[0];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);
      break;
    }
    case 2: {
      int blocks[1][2][2];
      int blockstride[2];

      blocks[0][0][0]=0;
      blocks[0][0][1]=0;

      blocks[0][1][0]=a.shape[0]-1;
      blocks[0][1][1]=a.shape[1]-1;

      blockstride[0] = a.stride[0];
      blockstride[1] = a.stride[1];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);

      break;
    }
    case 3: {
      int blocks[1][2][3];
      int blockstride[3];

      blocks[0][0][0]=0;
      blocks[0][0][1]=0;
      blocks[0][0][2]=0;

      blocks[0][1][0]=a.shape[0]-1;
      blocks[0][1][1]=a.shape[1]-1;
      blocks[0][1][2]=a.shape[2]-1;

      blockstride[0] = a.stride[0];
      blockstride[1] = a.stride[1];
      blockstride[2] = a.stride[2];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);

      break;
    }
    case 4: {
      int blocks[1][2][4];
      int blockstride[4];

      blocks[0][0][0]=0;
      blocks[0][0][1]=0;
      blocks[0][0][2]=0;
      blocks[0][0][3]=0;

      blocks[0][1][0]=a.shape[0]-1;
      blocks[0][1][1]=a.shape[1]-1;
      blocks[0][1][2]=a.shape[2]-1;
      blocks[0][1][3]=a.shape[3]-1;

      blockstride[0] = a.stride[0];
      blockstride[1] = a.stride[1];
      blockstride[2] = a.stride[2];
      blockstride[3] = a.stride[3];

      desc=create_f90_bdecomp_desc(a.rank,&blocks[0][0][0],tasks,1, blockstride, IC_EndPoint::IC_COLUMN_MAJOR);

      break;
    }

    default:
      assert(0);
  }

  return desc;
}')

define(`F90_ARRAY_REGISTER_HELPER',`
void IC_EndPointSet::registerArray(const char* ptname, const F90_$1Array& arr, int& stat){
    #ifdef IC_DEBUG
      cout << "Registering Port " << ptname << " with actual array." << endl;
    #endif

    string portname = ptname;
    stat = IC_ARRAY_REGISTER_FAILURE;        

    map<string, const F90_$1Array*>::const_iterator iter;
    iter = this->f90_$1_arrmap.find(portname);

    if(iter == this->f90_$1_arrmap.end()){
        #ifdef IC_DEBUG
            cout << "\tno such key in f90_$1_arrmap..." << endl;
        #endif
        stat = IC_ARRAY_REGISTER_FAILURE;
        return;
    }
    else{
        /* bind given array into the array map        */
        #ifdef IC_DEBUG
            cout << "\tportname found in f90_$1_arrmap..." << endl;
        #endif
        this->f90_$1_arrmap[portname] = &arr;
        stat = IC_OK;
    }

    if(stat == IC_OK){
        unsigned ndim = arr.rank;
        
        if (ndim > IC_MAX_DIM) {
            #ifdef IC_DEBUG
                cout << "\t" << IC_EndPoint::IC_errors[-IC_EndPoint::IC_INVALID_NDIM] << endl;
            #endif
            stat = IC_EndPoint::IC_INVALID_NDIM;
            return;
        }

        unsigned i;
        int lower[IC_MAX_DIM], upper[IC_MAX_DIM], stride[IC_MAX_DIM];
        
        for(i=0;i<ndim;++i) {
            lower[i] = 0; 
            upper[i] = arr.shape[i]-1; 
            stride[i] = 1;

#ifdef IC_DEBUG
            cout << "\tF90Array Structure: shape [" << i << "] " << arr.shape[i] << " stride[" << i << "] " << arr.stride[i] << endl;
#endif
        }

#ifdef IC_DEBUG
        cout << "\tnumber of dimensions is " << ndim << endl;
#endif

        map<string, IC_Connection*>::const_iterator citer;
        for (citer = this->connmap.begin(); citer != this->connmap.end(); ++citer) {
            string connname = citer->first;    

            IC_Connection* conn = (IC_Connection*)citer->second;
#ifdef IC_DEBUG
            cout << "\tCurrent..." << connname << " : "
                 << conn->get_portname() << endl;
#endif              
            
            if(!strcmp(conn->get_ctype().c_str(), "1xN")){
#ifdef IC_DEBUG
              cout << "\tSkipping 1xN connection: " << connname << endl;
#endif              
              continue;
            }

            if(conn->get_portname().compare(ptname) != 0){
#ifdef IC_DEBUG
              cout << "\tSkipping port" << endl;
#endif              
              continue;
            }

            /* create region*/
            IC_Region** region_set = (IC_Region**)malloc(sizeof(IC_Region*));
            region_set[0] = IC_Create_block_region(ndim, lower, upper, stride);
            if (!region_set[0]) {
              #ifdef IC_DEBUG
                cout << "\t" << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_ALLOC_REGION] << endl;
              #endif

              stat = IC_EndPoint::IC_CANT_ALLOC_REGION;
              return;
            }
            else{
         	    conn->set_rgn(region_set);
            }
	
            IC_EndPoint* ep = (IC_EndPoint*)conn->get_ep();
            unsigned myao = ep->get_myprog_arrayorder();
            unsigned oao  = ep->get_otherprog_arrayorder();
            #ifdef IC_DEBUG
                cout << "\tArray Order : " << myao << " " << oao <<endl;
            #endif
            
            IC_Desc* desc = IC_EndPoint::buildDescriptor(arr);
            if (!desc) {
                #ifdef IC_DEBUG
                    cout << "\t" << IC_EndPoint::IC_errors[-IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR] << endl;
                #endif
                stat = IC_EndPoint::IC_CANT_GET_DA_DESCRIPTOR;
                return;
            }
            else{
                conn->set_desc(desc);
            }
        }        
    }
    else{
        stat = IC_ARRAY_REGISTER_FAILURE;        
    }
}'
)






/*/// instantiatiated code*/
divert
#include "InterComm.h"
#include "Descriptor.h"
#include "IC_EndPoint.h"
#include "IC_EndPointSet.h"
#include <assert.h>
#include <iostream>

using namespace std;

#ifdef PPLUSPLUS
#include <A++.h>
#include "pplus2bparti.h"
#endif

#ifdef PPLUSPLUS

/* Methods for int arrays*/
PPP_EXPORT_HELPER(int)
PPP_IMPORT_HELPER(int)
PPP_BCAST_LOCAL_HELPER(int)
PPP_RECV_LOCAL_HELPER(int)
PPP_ICXJD_BCAST_LOCAL_HELPER(int)
PPP_ICXJD_RECV_LOCAL_HELPER(int)

/* Methods for float arrays*/
PPP_EXPORT_HELPER(float)
PPP_IMPORT_HELPER(float)
PPP_BCAST_LOCAL_HELPER(float)
PPP_RECV_LOCAL_HELPER(float)
PPP_ICXJD_BCAST_LOCAL_HELPER(float)
PPP_ICXJD_RECV_LOCAL_HELPER(float)

/* Methods for double arrays*/
PPP_EXPORT_HELPER(double)
PPP_IMPORT_HELPER(double)
PPP_BCAST_LOCAL_HELPER(double)
PPP_RECV_LOCAL_HELPER(double)
PPP_ICXJD_BCAST_LOCAL_HELPER(double)
PPP_ICXJD_RECV_LOCAL_HELPER(double)

/* Methods for arrays registration*/
PPP_ARRAY_REGISTER_HELPER(int)
PPP_ARRAY_REGISTER_HELPER(float)
PPP_ARRAY_REGISTER_HELPER(double)

#endif

#ifdef FORTRAN90

/* Methods for char arrays*/
F90_EXPORT_HELPER(char)
F90_IMPORT_HELPER(char)
F90_BCAST_LOCAL_HELPER(char)
F90_RECV_LOCAL_HELPER(char)
F90_ICXJD_BCAST_LOCAL_HELPER(char)
F90_ICXJD_RECV_LOCAL_HELPER(char)
F90_BUILD_DESCRIPTOR_HELPER(char)
F90_ARRAY_REGISTER_HELPER(char)

/* Methods for short arrays*/
F90_EXPORT_HELPER(short)
F90_IMPORT_HELPER(short)
F90_BCAST_LOCAL_HELPER(short)
F90_RECV_LOCAL_HELPER(short)
F90_ICXJD_BCAST_LOCAL_HELPER(short)
F90_ICXJD_RECV_LOCAL_HELPER(short)
F90_BUILD_DESCRIPTOR_HELPER(short)
F90_ARRAY_REGISTER_HELPER(short)

/* Methods for int arrays*/
F90_EXPORT_HELPER(int)
F90_IMPORT_HELPER(int)
F90_BCAST_LOCAL_HELPER(int)
F90_RECV_LOCAL_HELPER(int)
F90_ICXJD_BCAST_LOCAL_HELPER(int)
F90_ICXJD_RECV_LOCAL_HELPER(int)
F90_BUILD_DESCRIPTOR_HELPER(int)
F90_ARRAY_REGISTER_HELPER(int)

/* Methods for float arrays*/
F90_EXPORT_HELPER(float)
F90_IMPORT_HELPER(float)
F90_BCAST_LOCAL_HELPER(float)
F90_RECV_LOCAL_HELPER(float)
F90_ICXJD_BCAST_LOCAL_HELPER(float)
F90_ICXJD_RECV_LOCAL_HELPER(float)
F90_BUILD_DESCRIPTOR_HELPER(float)
F90_ARRAY_REGISTER_HELPER(float)

/* Methods for double arrays*/
F90_EXPORT_HELPER(double)
F90_IMPORT_HELPER(double)
F90_BCAST_LOCAL_HELPER(double)
F90_RECV_LOCAL_HELPER(double)
F90_ICXJD_BCAST_LOCAL_HELPER(double)
F90_ICXJD_RECV_LOCAL_HELPER(double)
F90_BUILD_DESCRIPTOR_HELPER(double)
F90_ARRAY_REGISTER_HELPER(double)

#endif
