/**
 * $Author: normanlo $
 * $Date: 2010/08/24 19:19:58 $
 * $Revision: 1.10 $
 * $Source: /fs/cvs/chaos/InterComm/src/xjd_handle.c,v $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "intercomm.h"
#include "pgme.h"
#include "ezxml/ezxml.h"
#include "xjd_handle.h"  
#include "general_bsparti.h"

int XJD_Validate_1_5(IC_XJD* xjd){
  if(check_multi_import(xjd))
    return -1;

   /** 
    * TODO: should implement later...
    *   inspect for the component and connection information 
    *   based on the version information
    */

   return 0; 
}
  
int check_multi_import(IC_XJD* xjd){
  ezxml_t curConn, tmpConn, exp, imp;
  char* impname;
  char* impport;
  char* tmp_impname;
  char* tmp_impport;
  int n = 0;
    
  curConn = get_first_conn(xjd->xmlRoot);
  while(curConn){
    impname = trim(ezxml_child(curConn, IMPORTER)->txt);
    impport = trim(ezxml_child(curConn, IMPPORT)->txt);
      
    tmpConn = get_first_conn(xjd->xmlRoot);
    while(tmpConn){
      tmp_impname = trim(ezxml_child(tmpConn, IMPORTER)->txt);
      tmp_impport = trim(ezxml_child(tmpConn, IMPPORT)->txt);
        
      if(!strcmp(impname, tmp_impname) && !strcmp(impport, tmp_impport)){
        n++;
      }
        
      tmpConn = tmpConn->next;
      free(tmp_impname);
      free(tmp_impport);
    }
      
    if(n > 1){
      fprintf(stderr, "\tImport from multiple program is detected.\n");
      return 1;
    }
    else{
      n = 0;
      curConn = curConn->next;
    }
    free(impname);
    free(impport);
  }

  return 0;
}


int XJD_Validate_2_0(IC_XJD* xjd){
   return 1; 
}

IC_XJD* IC_Initialize(char* pName, int rank, char* xjdname, int* icsts){
    IC_XJD* xjd = NULL;
    *icsts = -1;             // Assume failure

    pName = trim(pName);
    xjdname = trim(xjdname);

    xjd = (IC_XJD*) malloc(sizeof(IC_XJD));
    xjd->version = NULL;
    xjd->cHeader = NULL;
    xjd->cTail   = NULL;
    xjd->rHeader = NULL;
    xjd->rTail   = NULL;
    xjd->meou    = NULL;

    /* assign root of global XJD structure */
    xjd->xmlRoot = read_xjd(xjdname);
    if(xjd->xmlRoot == NULL){
      fprintf(stderr, "invalide XJD format. Parsing error.\n");
      free(xjd);
      return NULL;
    }
    
    /* get version info */
    if (get_version(xjd->xmlRoot) == NULL){
      fprintf(stderr, "XJD version is missing...\n");
      free_xjd(xjd->xmlRoot);
      free(xjd);
      free(pName);
      free(xjdname);
      return NULL;
    }
    else{
      xjd->version = strdup(get_version(xjd->xmlRoot));
    }
    /*fprintf(stderr, "version %s\n", xjd->version);*/

    if(!strcmp(xjd->version, "1.5")){
      if(XJD_Validate_1_5(xjd) < 0){
	free_xjd(xjd->xmlRoot);
	free(xjd->version);
	free(xjd);
	free(pName);
	free(xjdname);
        return NULL;
      }
      IC_Initialize_1_5(xjd, pName, rank, icsts);
    }
    else if(!strcmp(xjd->version, "2.0")){
      fprintf(stderr, "Version 2.0 is not supported\n");
      free_xjd(xjd->xmlRoot);
      free(xjd->version);
      free(xjd);
      free(pName);
      free(xjdname);
      return NULL;
    }
    else{
      fprintf(stderr, "Wrong version specified in XJD\n");
      free_xjd(xjd->xmlRoot);
      free(xjd->version);
      free(xjd);
      free(pName);
      free(xjdname);
      return NULL;
    }
  
    //*icsts = ((xjd == NULL) ? -1 : 0);
    if (*icsts < 0){
      free_xjd(xjd->xmlRoot);
      xjd->cHeader = xjd->cTail = NULL;
      xjd->rHeader = xjd->rTail = NULL;
      free(xjd->version);
      free(xjd);
      free(pName);
      free(xjdname);
      xjd = NULL;
    }

    free(pName);
    free(xjdname);
 
    return xjd;
}

void IC_Initialize_1_5(IC_XJD* xjd, char* pName, int rank, int* icsts){
    IC_XJDComponent* tmp;
    IC_XJDRegion* rgn;

    pName = trim(pName);

    /* register myself */
    *icsts = register_program(pName, xjd);
    if(*icsts < 0){
      //xjd = NULL;
      return;
    }

    /* inspect connections to look for the components related to me */
    *icsts = inspect_connections(pName, xjd);
    if(*icsts < 0){
      //xjd = NULL;
      return;
    }

    /* initializate components - set IC_Program ptr in IC_XJDComponent */
    tmp = (IC_XJDComponent*) xjd->cHeader;
    rgn = (IC_XJDRegion*) xjd->rHeader;

    tmp->comp = IC_Init(tmp->name, tmp->ntask, rank);
    tmp = (IC_XJDComponent*) tmp->next;

    fprintf(stderr, "program %s(rank %d) is IC_Waiting remote program\n", pName,rank);
    while(tmp){
      tmp->comp = IC_Wait(tmp->name, tmp->ntask);
      tmp = (IC_XJDComponent*) tmp->next;
    }

    /* initialize connections (regions) - connect IC_Programs in region */
    while(rgn){
        rgn->me = search_program(rgn->myname, xjd);  
        rgn->partner = search_program(rgn->partnername, xjd);
        
        /* to prevent partner terminates too early...
           not needed since we have IC_Sync at IC_Finalize, and IC_Wait before 
           handles waiting the other program.
           IC_Sync(rgn->me, rgn->partner);
        */
        
        rgn = (IC_XJDRegion*) rgn->next;
    }

    free(pName);

    fprintf(stderr, "======= IC_Initialize was successful. =======\n");
}

void IC_Initialize_2_0(){
    
}

/*
IC_RegionHandle* IC_InitHandle(IC_XJD* xjd, int* icsts){
    
    *icsts = 0;
    
    if(!strcmp(xjd->version, "1.5")){
        return (IC_RegionHandle*)IC_InitHandle_1_5();
    }
    else if(!strcmp(xjd->version, "2.0")){
        return (IC_RegionHandle*)IC_InitHandle_2_0();
    }
  
    *icsts = -1;
    return NULL;
}


IC_RegionHandle* IC_InitHandle_1_5(){
    IC_RegionHandle* handle = (IC_RegionHandle*) malloc(sizeof(IC_RegionHandle));
    handle->rgn = NULL;
    
    return handle;
}


IC_RegionHandle* IC_InitHandle_2_0(){
    IC_RegionHandle* handle = (IC_RegionHandle*) malloc(sizeof(IC_RegionHandle));

    //code here for SML 
        
    return handle;
}
*/

void IC_Register_region(IC_XJD* xjd, IC_Region** rgnset, int set_size, char* set_name, IC_Desc* desc, void* local_data, int* icsts){
    set_name = trim(set_name);


    if(!strcmp(xjd->version, "1.5")){
        IC_Register_region_1_5(xjd, rgnset, set_size, set_name, \
                               desc, local_data, icsts);
    }
    else if(!strcmp(xjd->version, "2.0")){
        IC_Register_region_2_0(xjd, rgnset, set_size, set_name, \
                               desc, local_data, icsts);
    }
    free(set_name);
}


void IC_Register_region_1_5(IC_XJD* xjd, IC_Region** rgnset, int set_size, char* set_name, IC_Desc* desc, void* local_data, int* icsts){
    IC_XJDRegion* rgn;
    IC_XJDRegion* searchptr = NULL;
    IC_Region** regioncopy;
    IC_XJDComponent* me = NULL;
    int i;
    int myao = -1; 
    int desc_ao = -1;

    rgn = search_region(set_name, xjd, -1, searchptr);

    while(rgn != NULL){
      /* useless ?
      if(rgn == NULL){
        fprintf(stderr, "%s is not used\n", set_name);
        *icsts = XJDREGION_REGISTER_ERR;
        return;
      }
      */

      /* check array order of my program and array order used for the desc */
      me = (IC_XJDComponent*) search_component(rgn->myname, xjd);
      if(me != NULL){
        myao = me->arrayorder;
      }
      else{
        /* this is impossible */
      }

      if(desc->type == IC_BDECOMP){
        desc_ao = ((decomp_Birreg*)desc->spec)->col_majeur;
      }
#ifdef IC_DEBUG      
      fprintf(stderr, "MyAO: %d, DescAO: %d\n", myao, desc_ao);
#endif
      if(myao != -1 && desc_ao != -1){
        if(myao != desc_ao){
          fprintf(stderr, "Component Array order is different to the array order used to build data descriptor.\n");
          //exit(1);
	  *icsts = -1;
	  return;
        }
      }

      regioncopy = (IC_Region**)calloc(set_size, sizeof(IC_Region*));
      for (i = 0; i < set_size; ++i) {
        regioncopy[i] = rgnset[i];
      }
      
      rgn->local_data = local_data;
      rgn->rgn_desc = desc;
      rgn->region = regioncopy;
      rgn->rgn_size = set_size;
      
      // set next search point
      searchptr = rgn->next;
      if(searchptr != NULL){
        rgn = search_region(set_name, xjd, -1, searchptr);
        if(rgn != NULL)
          searchptr = rgn->next;
      }
      else{
        rgn = NULL;
      }
    }
    
    *icsts = XJDREGION_REGISTER_OK;
}


void IC_Register_region_2_0(IC_XJD* xjd, IC_Region** rgnset, int set_size, char* set_name, IC_Desc* desc, void* local_data, int* icsts){
/*
    SMLregister("Ap2Sr1",  <=> set_name
                desc, <=> desc
                Sr1, <=> rgnset
                1,  <=> set_size
                local_data, <=> local_data
                sizeof(local_data), <=> Not exist.. We should resolve this.
                IC_float, <=> can find in XJD using set_name (in string)
                &th1 <=> handle
               )
*/                
}


void IC_Commit_region(IC_XJD* xjd, int* icsts){
    if(!strcmp(xjd->version, "1.5")){
        IC_Commit_region_1_5(xjd, icsts);
    }
    else if(!strcmp(xjd->version, "2.0")){
        IC_Commit_region_2_0();
    }
}


void IC_Commit_region_1_5(IC_XJD* xjd, int* icsts){
    IC_XJDRegion* rgn = (IC_XJDRegion*) xjd->rHeader;
    
    while(rgn){
        if(!strcmp(rgn->commtype, "MxN")){
            rgn->sched = IC_Compute_schedule(rgn->me, 
                                             rgn->partner, 
                                             rgn->rgn_desc,
                                             rgn->region,
                                             rgn->rgn_size);
            
            if(rgn->sched == NULL){
                fprintf(stderr, "sched generation for rgn with tag %d failed\n", rgn->tag);
                *icsts = IC_COMPUTE_SCHED_FAILED;
                return;
            }
        }
        rgn = (IC_XJDRegion*) rgn->next;
    }
    
#ifdef IC_DEBUG
    fprintf(stderr, "sched generation for all regions\n");
#endif
    *icsts = XJDREGION_COMMIT_OK;
}


void IC_Commit_region_2_0(){

}

int IC_Export(IC_XJD* xjd, char* rgnset_name){
    rgnset_name = trim(rgnset_name);
    if(!strcmp(xjd->version, "1.5")){
        return IC_Export_1_5(xjd, rgnset_name);
    }
    else if(!strcmp(xjd->version, "2.0")){
        return IC_Export_2_0(rgnset_name);
    }
    free(rgnset_name);
}  

int IC_Export_1_5(IC_XJD* xjd, char* rgnset_name){
    int status = -1;
    IC_XJDRegion* searchptr = NULL;
    IC_XJDRegion* rgn = search_region(rgnset_name, xjd, 0, searchptr);

    while(rgn != NULL){
       if(!strcmp(rgn->commtype, "MxN")){
           status = mxn_export_1_5(rgn);
       }
       
       if(status == -1){
           fprintf(stderr, "%s send with tag %d failed\n", rgn->my_region_name, rgn->tag);
           return IC_EXPORT_FAILED;
       }
       else{
           fprintf(stderr, "%s send with tag %d succeed\n", rgn->my_region_name, rgn->tag);
       }

       // set next search point
       searchptr = rgn->next;
       if(searchptr != NULL){
          rgn = search_region(rgnset_name, xjd, 0, searchptr);
          if(rgn != NULL)
             searchptr = rgn->next;
       }
       else{
          rgn = NULL;
       }
    }

    return IC_OK;
}

int IC_Export_2_0(char* rgnset_name){
    return IC_OK; 
}

int IC_Import(IC_XJD* xjd, char* rgnset_name){
    rgnset_name = trim(rgnset_name);

    if(!strcmp(xjd->version, "1.5")){
        return IC_Import_1_5(xjd, rgnset_name);
    }
    else if(!strcmp(xjd->version, "2.0")){
        return IC_Import_2_0(rgnset_name);
    }
    free(rgnset_name);
}  

/*
 * Note that a region name must be used as import only once in a XJD file
 */
int IC_Import_1_5(IC_XJD* xjd, char* rgnset_name){
    IC_XJDRegion* rgn = search_region(rgnset_name, xjd, 1, NULL);
    int status = -1;

    if(rgn == NULL){
      fprintf(stderr, "%s is not specified in the XJD.\n", rgnset_name);
      return IC_IMPORT_FAILED;
    }

    if(!strcmp(rgn->commtype, "MxN")){
        status = mxn_import_1_5(rgn);
    }

    if(status == -1){
        fprintf(stderr, "%s recv with tag %d failed \n", rgn->my_region_name, rgn->tag);
        return IC_IMPORT_FAILED;
    }
    else{
        fprintf(stderr, "%s recv with tag %d succeed\n", rgn->my_region_name, rgn->tag);
    }
        
    return IC_OK;
}


int IC_Import_2_0(char* rgnset_name){
    return IC_OK; 
}


/**
 * MxN region export
 * - long type is defined in the intercomm.h. However, it is not defined 
 * in the move.m4 file. 
 */
int mxn_export_1_5(IC_XJDRegion* rgn){
  int status = -1;

  if( !strcmp(rgn->type, "char")){
      status = IC_Send_char(rgn->partner, rgn->sched, (char*)rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "short")){
      status = IC_Send_short(rgn->partner, rgn->sched, (short*)rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "int")){
      status = IC_Send_int(rgn->partner, rgn->sched, (int*)rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "float")){
      status = IC_Send_float(rgn->partner, rgn->sched, (float*) rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "double")){
      status = IC_Send_double(rgn->partner, rgn->sched, (double*) rgn->local_data, rgn->tag);
  }

  return status;
}

/**
 * MxN region import
 * - long type is defined in the intercomm.h. However, it is not defined 
 * in the move.m4 file. 
 */
int mxn_import_1_5(IC_XJDRegion* rgn){
  int status = -1;

  if( !strcmp(rgn->type, "char")){
      status = IC_Recv_char(rgn->partner, rgn->sched, (char*)rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "short")){
      status = IC_Recv_short(rgn->partner, rgn->sched, (short*)rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "int")){
      status = IC_Recv_int(rgn->partner, rgn->sched, (int*)rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "float")){
      status = IC_Recv_float(rgn->partner, rgn->sched, (float*) rgn->local_data, rgn->tag);
  }
  else if( !strcmp(rgn->type, "double")){
      status = IC_Recv_double(rgn->partner, rgn->sched, (double*) rgn->local_data, rgn->tag);
  }

  return status;
}

/** 
 * <iyoon>
 * In fact, this functionality can be implemented with IC_Export 
 * by separating usual export and broadcast using 'commtype'. 
 * However, explicit representation by function name is better.
 *
 * The reason data and nelems are provided, is to give users 
 * flexibility to broadcast data to destination processes.
 * In case of 1xN region, the tag, data type and communication type 
 * are the only important information in region definition.
 * 1xN region is excluded in schedule generation 
 */
int IC_Bcast_local(IC_XJD* xjd, char* rgnset_name, void* data, int nelems){
    rgnset_name = trim(rgnset_name);
    return IC_Bcast_local_1_5(xjd, rgnset_name, data, nelems);
}



int IC_Bcast_local_1_5(IC_XJD* xjd, char* rgnset_name, void* data, int nelems){
    int status = -1;
    IC_XJDRegion* searchptr = NULL;
    IC_XJDRegion* rgn = search_region(rgnset_name, xjd, 0, searchptr);

    while(rgn != NULL){
       if(!strcmp(rgn->commtype, "1xN")){
           status = mxn_bcast_local_1_5(rgn, data, nelems);
       }
       
       if(status == -1){
           fprintf(stderr, "%s broadcast with tag %d failed\n", \
                   rgn->my_region_name, rgn->tag);
           return IC_EXPORT_FAILED;
       }
       else{
           fprintf(stderr, "%s broadcast with tag %d succeed\n", \
                   rgn->my_region_name, rgn->tag);
       }

       // set next search point
       searchptr = rgn->next;
       if(searchptr != NULL){
          rgn = search_region(rgnset_name, xjd, 0, searchptr);
          if(rgn != NULL)
             searchptr = rgn->next;
       }
       else{
          rgn = NULL;
       }
    }

    free(rgnset_name);

    return IC_OK;
}

/**
 * 1xN region export
 * - long type is defined in the intercomm.h. However, it is not defined 
 * in the move.m4 file. 
 */
int mxn_bcast_local_1_5(IC_XJDRegion* rgn, void* data, int nelems){
  int stat = -1;
  
  if( !strcmp(rgn->type, "char")){
    stat = IC_Bcast_local_char(rgn->partner, (char*)data, nelems, rgn->tag); 
  }
  else if( !strcmp(rgn->type, "short")){
    stat = IC_Bcast_local_short(rgn->partner, (short*)data, nelems, rgn->tag); 
  }
  else if( !strcmp(rgn->type, "int")){
    stat = IC_Bcast_local_int(rgn->partner, (int*)data, nelems, rgn->tag); 
  }
  else if( !strcmp(rgn->type, "float")){
    stat = IC_Bcast_local_float(rgn->partner, (float*)data, nelems, rgn->tag); 
  }
  else if( !strcmp(rgn->type, "double")){
    stat = IC_Bcast_local_double(rgn->partner, (double*)data, nelems, rgn->tag); 
  }
  
  return stat;
}


int IC_Recv_local(IC_XJD* xjd, char* rgnset_name, void* data, int nelems){
    rgnset_name = trim(rgnset_name);

    return IC_Recv_local_1_5(xjd, rgnset_name, data, nelems);
}

int IC_Recv_local_1_5(IC_XJD* xjd, char* rgnset_name, void* data, int nelems){
    IC_XJDRegion* rgn = search_region(rgnset_name, xjd, 1, NULL);
    int stat = 0;

    if(rgn == NULL){
      fprintf(stderr, "%s is not specified in the XJD.\n", rgnset_name);
      return IC_IMPORT_FAILED;
    }

    if(!strcmp(rgn->commtype, "1xN")){
        if( !strcmp(rgn->type, "char")){
            stat = IC_Recv_local_char(rgn->partner, (char*)data, nelems, rgn->tag); 
        }
        else if( !strcmp(rgn->type, "short")){
            stat = IC_Recv_local_short(rgn->partner, (short*)data, nelems, rgn->tag); 
        }
        else if( !strcmp(rgn->type, "int")){
            stat = IC_Recv_local_int(rgn->partner, (int*)data, nelems, rgn->tag); 
        }
        else if( !strcmp(rgn->type, "float")){
            stat = IC_Recv_local_float(rgn->partner, (float*)data, nelems, rgn->tag); 
        }
        else if( !strcmp(rgn->type, "double")){
            stat = IC_Recv_local_double(rgn->partner, (double*)data, nelems, rgn->tag); 
        }
    }
    else{
        stat = -1;            
    }
    
    if(stat == -1){
        fprintf(stderr, "%s recv (matched to broadcast) with %d failed\n", \
                rgn->my_region_name, rgn->tag);
        return IC_IMPORT_FAILED;
    }
    else{
        fprintf(stderr, "%s recv (matched to broadcast) with %d succeed\n", \
                rgn->my_region_name, rgn->tag);
    }
    free(rgnset_name);
        
    return IC_OK;
}


/*
void IC_Free_RegionHandle(IC_XJD* xjd, IC_RegionHandle* handle, int* icsts){
    if(!strcmp(xjd->version, "1.5")){
        IC_Free_RegionHandle_1_5(handle, icsts);
    }
    else if(!strcmp(xjd->version, "2.0")){
        IC_Free_RegionHandle_2_0(handle, icsts);
    }
}


void IC_Free_RegionHandle_1_5(IC_RegionHandle* handle, int* icsts){
    *icsts = 0; 
    free(handle);  
}


void IC_Free_RegionHandle_2_0(IC_RegionHandle* handle, int* icsts){
    *icsts = 0;
}
*/    

void IC_Finalize(IC_XJD* xjd, int* icsts){
    if(!strcmp(xjd->version, "1.5")){
        IC_Finalize_1_5(xjd, icsts);
    }
    else if(!strcmp(xjd->version, "2.0")){
        IC_Finalize_2_0();
    }
}

    
void IC_Finalize_1_5(IC_XJD* xjd, int* icsts){
  IC_XJDRegion *ptr = NULL; 
  IC_XJDProgram *list = NULL;
  IC_XJDProgram *tmp = NULL;
  IC_XJDProgram *tmp_prev = NULL;
  IC_XJDProgram *entry = NULL;
  int i;
  

#ifdef IC_DEBUG
  fprintf(stderr, "======= Removing Region info. (Finalizing)=======\n");
#endif

  /* remove all IC_XJDRegion*/
  if(xjd->rHeader != NULL){
    /* sync programs to prevent early termination */
    //    ptr = (IC_XJDRegion*) xjd->rHeader;
    //    while(ptr != NULL){
    //      IC_Sync(ptr->me, ptr->partner);
    //      ptr = ptr->next;
    //    }

    /*Build the list of IC_Programs*/
      if ((entry = malloc (sizeof(IC_XJDProgram))) == NULL) {
	fprintf (stderr, "Error creating the list of programs\n");
	*icsts = -1;
	return;
      }
    /* Adding the first element */
    ptr = (IC_XJDRegion*) xjd->rHeader;
    if (ptr != NULL){
      /* No need to allocate extra spaces */
	entry->name = ptr->partnername;
	entry->program = ptr->partner;
	entry->next = NULL;
    }
    list = entry;

    /* Adding the rest of elements */
    ptr = ptr->next;
    while (ptr != NULL){
      tmp = list;
      while (tmp != NULL && strcmp(tmp->name, ptr->partnername) < 0)
        tmp = tmp->next;
      /* If the new entry is not a duplicate, add it to the list */
      if ((tmp != NULL && strcmp (tmp->name, ptr->partnername) != 0) || 
	   tmp == NULL){
        if ((entry = malloc (sizeof(IC_XJDProgram))) == NULL){
          fprintf (stderr, "Error creating the list of programs\n");
          //exit (-1);
	  *icsts = -1;
	  return;
	}
	entry->name = ptr->partnername;
        entry->program = ptr->partner;
        entry->next = NULL;


        /* find the previous pointer in the linked list */
        tmp_prev = list;
        if (tmp != list){
          while (tmp_prev->next != tmp)
	    tmp_prev = tmp_prev->next;
        }
        
        /* Inserting the new node into the list */
        if (tmp == list){
          entry->next = list;
          list = entry;
        }
        else if (tmp != NULL){
          entry->next = tmp_prev->next;
          tmp_prev->next = entry;
	}
        else{
          entry->next = NULL; 
          tmp_prev->next = entry;
        }
      }
      ptr = ptr->next;
    }  

    #ifdef IC_DEBUG
      i = 0;
      fprintf(stderr, "========== The programs that are going to be synced\n");
      tmp = list;
      while (tmp){
        fprintf (stderr, "%dth program, %s\n", i++, tmp->name);
        tmp = tmp->next;
      }
      fprintf(stderr, "=============================================\n");
    #endif

    /* Calling IC_Sync for each program in the list */
    ptr = (IC_XJDRegion*) xjd->rHeader;
    tmp = list;
    while (tmp != NULL){
      IC_Sync(ptr->me, tmp->program);
      tmp = tmp->next;
    }

    /* release memory */
    release_rgn((IC_XJDRegion*) xjd->rHeader);
  }

#ifdef IC_DEBUG
  fprintf(stderr, "======= Removing Component info. (Finalizing)=======\n");
#endif

  /* remove all IC_XJDComponent*/
  if(xjd->cHeader != NULL){
    IC_XJDComponent* firstcomp = (IC_XJDComponent*) xjd->cHeader;
    
    release_comp( (IC_XJDComponent*) firstcomp->next, 0);
    release_comp( (IC_XJDComponent*) firstcomp, 1);
  }

  free_xjd(xjd->xmlRoot);
  
  xjd->cHeader = xjd->cTail = NULL;
  xjd->rHeader = xjd->rTail = NULL;
  free(xjd->version);
  free(xjd);
  xjd = NULL;
  
  fprintf(stderr, "======= IC_Finalize was successful. =======\n");

  *icsts = 0;
}    


void IC_Finalize_2_0(){

}    


void IC_Sync_by_rgn(char* rgn_name, IC_XJD* xjd, int* icsts){
    IC_XJDRegion* rgn = search_region(rgn_name, xjd, -1, NULL);
    IC_Sync(rgn->me, rgn->partner);    
    *icsts = 0;
}


void release_rgn(IC_XJDRegion* ptr){
    if(ptr == NULL)
        return;
    
    if(ptr->next != NULL)
        release_rgn((IC_XJDRegion*) ptr->next);

    /* sync before termination -- moved to IC_Finalize 
       IC_Sync(ptr->me, ptr->partner); */

    /* release memory for IC_XJDRegion*/
    if(ptr->sched){
        IC_Free_sched(ptr->sched);
    }
    if(ptr->region){
        free(ptr->region);
    }
    free(ptr->myname);
    free(ptr->partnername);
    free(ptr->type);
    free(ptr->commtype);
    free(ptr->my_region_name);
    free(ptr->partner_region_name);
    free(ptr);
}


void release_comp(IC_XJDComponent* ptr, int first){
    if(ptr == NULL){
        return;
    }

    if(first){
        IC_Quit(ptr->comp);
        free(ptr->name);
        free(ptr);
        return;
    }
    
    if(ptr->next != NULL){
        release_comp((IC_XJDComponent*) ptr->next, 0);
    }

    /* release memory for IC_XJDComponent*/
    IC_Free_program(ptr->comp);
    free(ptr->name);
    free(ptr);
}

int inspect_connections(const char* myname, IC_XJD* xjd){
    ezxml_t curConn, exp, imp;
    char* cid;
    char* pname;
    char* pname2;
    int sts;
    
    curConn = get_first_conn(xjd->xmlRoot);
    
    while(curConn){
        exp = ezxml_child(curConn, EXPORTER);
        pname = search_pname_byID(trim(exp->txt), xjd);
        
        if( !strcmp(myname, pname) ){
#ifdef IC_DEBUG
            fprintf(stderr, "\t%s appear as exp\n", pname);
#endif            
            imp = ezxml_child(curConn, IMPORTER);
            pname2 = search_pname_byID(trim(imp->txt), xjd);
            
            if( search_program(pname2, xjd) == NULL){
                sts = register_program(pname2, xjd);
                if( sts < 0 ){
                  return XJDCOMP_REGISTER_ERR;
                }
            }
            
            register_region( \
                    pname, pname2,  \
                    trim(ezxml_child(curConn, EXPPORT)->txt),	\
                    trim(ezxml_child(curConn, IMPPORT)->txt),	\
		    trim(ezxml_child(curConn, CONNID)->txt),	\
                    trim(ezxml_child(curConn, DATATYPE)->txt),	\
                    trim(ezxml_child(curConn, COMMTYPE)->txt),	\
                    atoi(trim(ezxml_child(curConn, TAG)->txt)),	\
                    0, \
                    xjd);
                    
        }
        
        imp = ezxml_child(curConn, IMPORTER);
        pname = search_pname_byID(trim(imp->txt), xjd);
        
        if( !strcmp(myname, pname) ){
#ifdef IC_DEBUG
            fprintf(stderr, "\t%s appear as importer\n", pname);
#endif
            exp = ezxml_child(curConn, EXPORTER);
            pname2 = search_pname_byID(trim(exp->txt), xjd);
            
            if( search_program(pname2, xjd) == NULL){
                sts = register_program(pname2, xjd);
                if( sts < 0 ){
                  return XJDCOMP_REGISTER_ERR;
                }
            }
            
            register_region( \
                    pname, pname2,  \
                    trim(ezxml_child(curConn, IMPPORT)->txt),	\
                    trim(ezxml_child(curConn, EXPPORT)->txt),	\
		    trim(ezxml_child(curConn, CONNID)->txt),	\
                    trim(ezxml_child(curConn, DATATYPE)->txt),	\
                    trim(ezxml_child(curConn, COMMTYPE)->txt),	\
                    atoi(trim(ezxml_child(curConn, TAG)->txt)),	\
                    1, \
                    xjd);
                    
        }
        
        exp = imp = NULL;
        curConn = curConn->next;
    }

    return IC_OK;
}


int register_program(const char* pName, IC_XJD* xjd){
    ezxml_t curComp;
    ezxml_t tmp, tmpstr, cid;
    IC_XJDComponent* aComp;
   
    curComp = get_first_comp(xjd->xmlRoot);
    
    while(curComp){
        cid = ezxml_child(curComp, COMPID);
        tmp = ezxml_child(curComp, COMPNAME);

        if(cid == NULL){
	  fprintf(stderr, "Missing identifier for component %s\n", pName);
            break;
        }
        if(tmp == NULL){
	  fprintf(stderr, "Unable to find component name %s in XJD...\n", pName);
            break;
        }
        
        if( !strcmp(pName, trim(tmp->txt)) ){
            aComp = (IC_XJDComponent*) malloc(sizeof(IC_XJDComponent));
           
            aComp->id = (char*)strdup(trim(cid->txt));
            aComp->name = (char*)strdup(pName);
            
            tmpstr = ezxml_child(curComp, N_NODE); 
            if(tmpstr){
              aComp->ntask = atoi(trim(tmpstr->txt));
            }
            else{
              fprintf(stderr, "nNode is not defined for component %s.\n", pName);
            }
            
            tmpstr = ezxml_child(curComp, ARRAYORDER);
            if(tmpstr){
              if(!strcmp(trim(tmpstr->txt), "IC_ROW_MAJOR")){
                aComp->arrayorder = IC_ROW_MAJOR;
              }
              else if(!strcmp(trim(tmpstr->txt), "IC_COLUMN_MAJOR")){
                aComp->arrayorder = IC_COLUMN_MAJOR;
              }
              else{
                fprintf(stderr, "invalid array order for the component %s\n", pName);
                return XJDCOMP_REGISTER_ERR;
              }
            }
            else{
              fprintf(stderr, "arrayorder for component %s must be specified.\n", pName);
              return XJDCOMP_REGISTER_ERR;
            }

            aComp->comp = (IC_Program*) NULL;
            aComp->next = (IC_XJDComponent*) NULL;
            
            add_comp((IC_XJDComponent*)aComp, xjd);
            
            return XJDCOMP_REGISTER_OK;
        }

        curComp = curComp->next;
    }
    
    return XJDCOMP_REGISTER_ERR;
}

int register_region(
    char* myname, char* partnername,
    char* my_region_name, char* partner_region_name, char *connid,
    char* type, char* commtype, int tag, int usetype, IC_XJD* xjd){
    IC_XJDRegion* aRegion;
            
#ifdef IC_DEBUG 
    fprintf(stderr, "Register Region %s\n", my_region_name);
#endif
    
    aRegion = (IC_XJDRegion*) malloc(sizeof(IC_XJDRegion));

    aRegion->myname = (char*) strdup(myname);
    aRegion->partnername = (char*) strdup(partnername);
    
    aRegion->my_region_name = (char*) strdup(my_region_name);
    aRegion->partner_region_name = (char*) strdup(partner_region_name);
    aRegion->connid = (char *) strdup (connid);
    aRegion->type = (char*) strdup(type);
    aRegion->commtype = (char*) strdup(commtype);
    
    aRegion->me = (IC_Program*) NULL;
    aRegion->partner = (IC_Program*) NULL;
    
    aRegion->tag = tag;
    aRegion->usetype = usetype;
    aRegion->region = (IC_Region**) NULL; /* filled later.*/
    aRegion->rgn_size = 0; /* filled later.*/
    
    aRegion->rgn_desc = (IC_Desc*) NULL; /* filled later.*/
    aRegion->sched = (IC_Sched*) NULL;   /* filled later.*/
    aRegion->local_data = (void*) NULL;  /* filled later.*/

    aRegion->next = (IC_XJDRegion*) NULL;
        
    add_region((IC_XJDRegion*) aRegion, xjd);

    return 0;
}
                                
char* search_pname_byID(char* id, IC_XJD* xjd){
    ezxml_t curComp;
    ezxml_t cid, tmp;
    IC_XJDComponent* aComp;
   
    curComp = get_first_comp(xjd->xmlRoot);
    
    while(curComp){
        cid = ezxml_child(curComp, COMPID);
        tmp = ezxml_child(curComp, COMPNAME);

        if( !strcmp(trim(cid->txt), id) ){
	  return trim(tmp->txt);
        }

        curComp = curComp->next;
    }

    return NULL;
}
   
IC_XJDComponent* search_component(char* name, IC_XJD* xjd){
    IC_XJDComponent* tmp = xjd->cHeader;
    
    while(tmp){
        if( !strcmp(tmp->name, name) ){
            return tmp;
        }
        tmp = tmp->next;
    }
    
    return NULL;
}

IC_Program* search_program(char* name, IC_XJD* xjd){
    IC_XJDComponent* tmp = xjd->cHeader;
    
    while(tmp){
        if( !strcmp(tmp->name, name) ){
            return tmp->comp;
        }
        tmp = tmp->next;
    }
    
    return NULL;
}


IC_XJDRegion* search_region(char* name, IC_XJD* xjd, int usetype, IC_XJDRegion* searchptr){
    IC_XJDRegion* cur = xjd->rHeader;
    
    if(searchptr != NULL){
       cur = searchptr;
    }
    
    while(cur != NULL){
        if( !strcmp(cur->my_region_name, name)){
           if(usetype == 0){ // export
              if(cur->usetype == 0){
                 return cur;
              }
           }
           else if(usetype == 1){ // import
              if(cur->usetype == 1){
                 return cur;
              }
           }
           else{
              return cur;
           }
        }

        cur = cur->next;
    }
    
    return NULL;
}


void add_comp(IC_XJDComponent* comp, IC_XJD* xjd){
    if( xjd->cHeader == NULL ){
        xjd->cHeader = comp;
        xjd->cTail   = comp;
    }
    else{
        xjd->cTail->next = comp;
        xjd->cTail = comp;
    }
}



void add_region(IC_XJDRegion* region, IC_XJD* xjd){
  /* Insert the region in ordered fashion so it will match */
  /* the list of region in high-level interface, which uses map */
  IC_XJDRegion *tmp;
  IC_XJDRegion *tmp_prev;
  
    /* Adding first element */
    if( xjd->rHeader == NULL){
        xjd->rHeader = region;
        xjd->rTail = region;
    }
    /* If the list in xjd is not empty */
    else{
        tmp = xjd->rHeader;
        /* Search the position of where the new entry should be inserted */
        while (tmp != NULL && strcmp (tmp->connid, region->connid) < 0)
          tmp = tmp->next;

        /* find the previous pointer in the linked list */
        tmp_prev = xjd->rHeader;
        if (tmp != xjd->rHeader) {
          while (tmp_prev->next != tmp)
            tmp_prev = tmp_prev->next;
	}

        /* Insert the new node into the list */
        if (tmp == xjd->rHeader){
          region->next = xjd->rHeader;
          xjd->rHeader = region;
	}
        else if (tmp != NULL){
          region->next = tmp_prev->next;
          tmp_prev->next = region;
	}
        else{
          region->next = NULL;
          tmp_prev->next = region;
          xjd->rTail = region;
	}
    }
}


void print_components(IC_XJD* xjd){
    IC_XJDComponent* tmp = (IC_XJDComponent*) xjd->cHeader;
    
    fprintf(stdout, "======= COMPONENT LIST =======\n");
    while(tmp){
        print_comp(tmp);
        tmp = (IC_XJDComponent*) tmp->next;
    }
}


void print_comp(IC_XJDComponent* tmp ){
    fprintf(stderr, "  Name: %s\n  Task: %d\n", \
                            tmp->name, tmp->ntask);
    if(tmp->comp){
        fprintf(stderr, "  Comp: %s\n\n", tmp->comp->name);
    }
    else{
        fprintf(stderr, "\n");
    }
}    
    

void print_regions(IC_XJD* xjd){
    int i;
    IC_XJDRegion* tmp = (IC_XJDRegion*) xjd->rHeader;
    
    fprintf(stdout, "========= REGION LIST =========\n");
    while(tmp){
        print_region(tmp);
        tmp = (IC_XJDRegion*) tmp->next;
    }
}

void print_region(IC_XJDRegion* tmp ){
    int i;
    
    if(tmp->me && tmp->partner){
        fprintf(stdout, "  Me: %s %d\n  Other: %s %d\n  MyRgn: %s\n  OtherRgn: %s\n  RgnType: %s\n", \
                            tmp->me->name, tmp->me->nbNode, \
                            tmp->partner->name, tmp->partner->nbNode, \
                            tmp->my_region_name, \
                            tmp->partner_region_name, \
                            tmp->type );
                                
        for ( i=0 ; i<tmp->rgn_size ; i++){
            fprintf(stdout, "  RgnSize: %d\n  RgnType: %d", tmp->region[i]->size, tmp->region[i]->type);
        }                                
    }
    else{
        fprintf(stdout, "  Me: %s \n  Other: %s \n  MyRgn: %s\n  OtherRgn: %s\n  RgnType: %s\n", \
                            tmp->myname, tmp->partnername, \
                            tmp->my_region_name, \
                            tmp->partner_region_name, \
                            tmp->type );
    }        
    fprintf(stdout, "\n");
}


int parse_param(int argc, char **argv, char** xjd){
    int c;
    int dflag = 0;

    optind = 1; /* reset option index to 1 to parse from the first option*/
   
    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'd':
                *xjd = strdup(optarg);
                dflag = 1;
                break;
        }
    }

    if(!dflag){
        return -1;
    }
    else{
        return 0;
    }
}

ezxml_t read_xjd(const char* xjdname){
    FILE* fp; 
   
    if ((fp = fopen(xjdname, "r")) == NULL){
        fprintf(stderr, "Cannot open %s\n", xjdname);
        return NULL;
    }
            
    return ezxml_parse_fp(fp);
}


void free_xjd(ezxml_t xmlRoot){ 
    ezxml_free(xmlRoot);
}

char* get_version(ezxml_t xmlRoot){
  ezxml_t v = ezxml_child(xmlRoot, XJDVERSION);
  if(v != NULL){
    return trim(v->txt);
  }
  else{
    return NULL;
  }
}

ezxml_t get_first_comp(ezxml_t xmlRoot){
    ezxml_t plist = ezxml_child(xmlRoot, COMPLIST);
    return ezxml_child(plist, COMP);
}

ezxml_t get_first_conn(ezxml_t xmlRoot){
    ezxml_t clist = ezxml_child(xmlRoot, CONNLIST);
    return ezxml_child(clist, CONN);
}

char *trim(char *str){
  char *end;
  char *newstr;

  /* Remove leading white spaces */
  while (isspace(*str))str++;

  /* Copy the new string into another memory location */
  newstr = malloc(sizeof(char)*(strlen(str)+1));
  strcpy(newstr, str);

  /* Remove Trailing white spaces */
  end = newstr + strlen(newstr) - 1;
  while (end > newstr && isspace(*end)) end--;
  *(end+1) = 0;

  return newstr;
}


