      program fexample
      implicit none
      include 'mpif.h'

      character*8 localname
      data localname / 'cexample' /
      character*8 othername
      data othername / 'fexample' /
      integer local_tasks
      data local_tasks / 4 /
      integer other_tasks
      data other_tasks / 8 /
      integer i, rank

      integer groupcomm
      
      integer local
      integer other

      integer desc
      real A(3,4,5)
      integer broadA(10)
      integer broadTag
      data broadTag / 77 /
      
      integer region_set(2)
      integer lower(3), upper(3), stride(3)
      integer sched
      integer tag
      data tag / 99 /

      integer sts

c     initialize mpi
      print *, 'initialize mpi'
      call MPI_INIT(sts)
      call MPI_COMM_DUP(MPI_COMM_WORLD, groupcomm, sts)
      call MPI_COMM_RANK(groupcomm, rank, sts)
      call MPI_BARRIER(groupcomm, sts)

c     initialize ic
      print *, 'initialize ic'
      call IC_Init(localname, local_tasks, rank, local)    
      call IC_Wait(othername, other_tasks, other)
      call IC_Sync(local, other, sts)

c     receiving broadcasting from the other program
      call IC_Recv_local_int(other, broadA, 10, broadTag, sts)
      print *, 'succesfully received broadcasting'
      print *, broadA

c     create a block decomposition type array descriptor and array
      call create_bdecomp(local_tasks, rank, A, desc)

c     define two 2x2x2 element regions for transfer
      print *, 'define regions'
      do i = 1, 3
         lower(i) = 1
         upper(i) = 1+i
         stride(i) = 1
      end do
      call IC_Create_block_region(3, lower, upper, stride,
     $     region_set(1))

      lower(1) = 1
      upper(1) = 2
      stride(1) = 1
      lower(2) = 5
      upper(2) = 7
      stride(2) = 1
      lower(3) = 6
      upper(3) = 9
      stride(3) = 1
      call IC_Create_block_region(3, lower, upper, stride,
     $     region_set(2))

      
      print *, 'create schedule'
      call IC_Compute_schedule(local, other, desc, region_set,
     $     2, sched)
      print *, 'send data'
      call IC_Send_float(other, sched, A, tag, sts)     
      call IC_Sync(local, other, sts)
      call IC_Free_sched(sched)
      
c     cleanup
      print *, 'clean up'
      call IC_Free_region(region_set(2))
      call IC_Free_region(region_set(1))

      call IC_Free_desc(desc)

      call IC_Free_program(other)      
      call IC_Quit(local, sts)
      
      call MPI_BARRIER(groupcomm, sts)
      call MPI_COMM_FREE(groupcomm)
      call MPI_FINALIZE(sts)

      end

      subroutine create_bdecomp(p, r, A, desc)
      implicit none
      integer p, r
      real A(3,4,5)
      integer desc

      integer blocks(4,2,3)
      integer tasks(4)
      integer i, j, k
      integer blkstr(3)

c     bisect the global array in each dimension (2x2 = 4 block)
      do i = 0, 1
         do j = 0, 1
            blocks(2*i+j+1,1,1) = 1
            blocks(2*i+j+1,1,2) = i*4 + 1
            blocks(2*i+j+1,1,3) = j*5 + 1

            blocks(2*i+j+1,2,1) = 3
            blocks(2*i+j+1,2,2) = i*4 + 4
            blocks(2*i+j+1,2,3) = j*5 + 5
         end do
      end do

c     assign the blocks to the tasks
      do i = 1, 4
         tasks(i) = i - 1
      end do

c     create the ic descriptor
      blkstr(1) = 1
      blkstr(2) = 3
      blkstr(3) = 12
      call IC_Create_bdecomp_desc(3, blocks, tasks, 4, desc, blkstr, 1)

c     assign each local element a number based on its index
      do i = 1, 3
         do j = 1, 4
            do k = 1, 5
               A(i,j,k) = 10000*r + 1000*i + 100*j + 10*k
            end do
         end do
      end do

      end



