/**
 * $Author: iyoon $
 * $Date: 2009/10/01 21:56:20 $
 * $Revision: 1.2 $
 * $Source: /fs/cvs/chaos/InterComm/test/multicast/pvm_tls.c,v $
 * 
 * Modified by Il-Chul Yoon <iyoon@cs.umd.edu>
 */


#include "pvm_tls.h"
#include <pvm3.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

/* initialize pvm */
int init_pvm(int argc, char** argv, char* groupName, int* nTask, int* rank){
  if ((*nTask = pvm_startup(argc, argv)) < 0){
    return -1;
  }
  
  if ((*rank = pvm_joingroup(groupName)) < 0){
    return -1;
  }
  
  pvm_barrier(groupName, *nTask);
  return *rank;
}


/* initialize pvm */
int init_pvm_mfile(int argc, char** argv, char* groupName, int* nTask, int* rank){

  if ((*nTask = start_master_slave(argc, argv)) < 0){
    return -1;
  }

  if ((*rank = pvm_joingroup(groupName)) < 0){
    return -1;
  }
  
  pvm_barrier(groupName, *nTask);
  return *rank;
}


/* start all processes on a node  */
int pvm_startup(int argc, char* argv[]) {
  int i;
  int parent;
  int started;
  int tids[64];
  int nTask = -1; 
  
  nTask = get_ntask(argc, argv);
  
  if (nTask < 2){
    return nTask;
  }
  
  parent = pvm_parent();
  if (parent < 0) {
    pvm_catchout(stdout);
    pvm_catchout(stderr);
    
    started = pvm_spawn(argv[0], argv+1, PvmTaskDefault, 0, nTask-1, tids);
    
    if (started < nTask-1) {
      pvm_perror("spawn error");
      for (i = 0; i < nTask-1; ++i) {
        if (tids[i] > 0) {
          pvm_kill(tids[i]);
        }
      }
      pvm_exit();
      return -1;
    } 
    else {
      ++started; /* include the parent process*/
    }
  } 
  else {
    started = nTask;
  }
  
  return started;  
}


/* start a process per a node */
int start_master_slave(int argc, char* argv[]) {
    int i, j;
    int tids[100];
    int erm;
    int parent;
    FILE* fp;
    char nodename[100]; /* restriction: the max. nodename to 100 characters*/
    char* mfile = NULL;
    int nTask = -1; 
      
    nTask = get_ntask(argc, argv);
    mfile = get_machinefile(argc, argv);
      
    if (nTask < 2){
        return nTask;
    }
      
    if(mfile == NULL){
        fprintf(stderr, "Machine file is not ready.\n");
        return -1;
    }


    if (pvm_parent() == PvmNoParent) {
	fprintf(stdout, "hello,world\n");        
        pvm_catchout(stdout);
        pvm_catchout(stderr);

        fprintf(stderr,"Spawning %d on %s hosts\n", nTask, mfile);
        fprintf(stderr,"Catching stdout and stderr\n");

        if( (fp = fopen(mfile, "r")) == NULL){
            fprintf(stderr, "Machine file open error\n");
            return -1;
        }

        if(fgets(nodename, 100, fp) == NULL){
            fprintf(stderr, "Nodename can't be retrieved from machine file\n");
            return -1;
        }
      
        i = 0;
        while(fgets(nodename, 100, fp) != NULL){
	  nodename[strlen(nodename)-1] = 0; /* cut newline character */
            fprintf(stderr, "name: <%s>\n", nodename);
    
            erm = pvm_spawn(argv[0], argv+1, PvmTaskHost, nodename, 1, &tids[i]);
            i++;
    
            if(erm != 1){
                pvm_perror("Error occurred in spawning...\n");
    
                for (j=2 ; j<nTask-1 ; j++){
                    if(tids[j] > 0){
                        pvm_kill(tids[j]);
                    }
                }
                pvm_exit();
                return -1;
            }
        }
        
        fclose(fp);

        pvm_initsend(PvmDataDefault);
        pvm_pkint(&nTask, 1, 1);

        if (erm = pvm_mcast(tids, nTask-1, 001) < 0){
            pvm_perror("error in first mcast\n");
        }
        fprintf(stderr, "\nStarted %i tasks (one for each node)\n", nTask);
    }
    else {
        parent = pvm_parent();
        fprintf(stderr, "parent detected\n");
        pvm_recv(parent, 001);
        pvm_upkint(&nTask, 1, 1);
        sleep(2);
    }

    return nTask;
}


int pvm_dupgroup(char* groupname, int r) {
  if (r == 0) {
    return pvm_joingroup(groupname);
  } else {
    while (pvm_gsize(groupname) != r)
      sleep(1);
    return pvm_joingroup(groupname);
  }
}

int pvm_allstatus(char* groupname, int status, int msgtag) {
  int allstatus;
  int i, mytid, gsize;

  mytid = pvm_mytid();
  gsize = pvm_gsize(groupname);

  pvm_initsend(PvmDataDefault);
  pvm_pkint(&status, 1, 1);
  pvm_bcast(groupname, msgtag);

  allstatus = status;
  for (i = 0; i < gsize; ++i) {
    if (mytid != pvm_gettid(groupname, i)) {
      pvm_recv(pvm_gettid(groupname, i), msgtag);
      pvm_upkint(&status, 1, 1);
      allstatus *= status;
    }
  }

  pvm_barrier(groupname, gsize);

  return allstatus;
}


int get_ntask(int argc, char **argv){
    int c;
    int nflag = 0;
    int ntask = -1;

    optind = 1; /* reset option index to 1 to parse from the first option*/

    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'n':
                ntask = atoi(optarg);
                nflag = 1;
                break;
        }
        /*printf("%d %d\n", c, nflag);*/
    }

    if(!nflag){
      return -1;
    }
    else{
      return ntask;
    }
}


char* get_machinefile(int argc, char **argv){
    int c;
    int mflag = 0;
    char* mfile ; 
    
    optind = 1; /* reset option index to 1 to parse from the first option*/

    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'm':
                mfile = optarg;
                mflag = 1;
                break;
        }
    }

    if(!mflag){
      return NULL;
    }
    else{
      return mfile;
    }
}
