/**
 * $Author: normanlo $
 * $Date: 2010/09/13 23:21:02 $
 * $Revision: 1.2 $
 * $Source: /fs/cvs/chaos/InterComm/test/xjd-ttable/recv2.c,v $
 * 
 * Written by Il-Chul Yoon <iyoon@cs.umd.edu>
 */

#include <intercomm.h>
#include <mpi.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#define COMPONENT_NAME "recv2comp"
#define REGION_NAME1   "recv2R1"

/**
 * Local view
 * Initialize 1-dim float array
 */
void init_data(int rank, float data[5]) {
    int i;
    
    fprintf(stderr, "initialize data\n");
    for (i = 0 ;  i < 5 ; i++){
      data[i] = -1.0;
    }
}

void print_array(int rank, float data[5]){
    int i;
    
    fprintf(stdout, "My rank is : %d\n", rank);
    for( i = 0 ; i < 5 ; i++ ){
      fprintf(stdout, "%4.0f  ", data[i]);
    }
    fprintf(stdout, "\n");
}

int parse_xjdname(int argc, char** argv, char** xjdname){
    int c; 
    int xjdflag = 0;
    
    optind = 1; /* reset option index to 1 to parse from the first option*/
 
    while ((c = getopt (argc, argv, "m:n:d:")) != -1){
        switch(c){
            case 'd':
                *xjdname = (char*)strdup(optarg);
    
                xjdflag = 1;
                break;
        }
    }
    
    if(!xjdflag){
        return -1;
    }
    else{
        return 0;
    }
}   
    

/**
 * Global view
 * This part should be decided by application programmer
 */
IC_Desc *create_block() {
    int blocks[2][2][1] = {
        { {0}, {3} },
        { {4}, {9} },
    };
    
    int tasks[2]= {0, 1};
    
    return IC_Create_bdecomp_desc(1, &blocks[0][0][0], tasks, 2, IC_ROW_MAJOR);
}

/**
 * This part should be decided by application programmer
 */
void create_region1(IC_Region* region[2]){
    int lower[1], upper[1], stride[1];

    lower[0] = 0;
    upper[0] = 1;
    stride[0] = 1;
    region[0] = IC_Create_block_region(1, lower, upper, stride);
    
    lower[0] = 6;
    upper[0] = 7;
    stride[0] = 1;
    region[1] = IC_Create_block_region(1, lower, upper, stride);
}

void initMPI(int* argc, char*** argv, int* rank, MPI_Comm* groupcomm)
{
    MPI_Init(argc, argv);
    MPI_Comm_dup(MPI_COMM_WORLD, groupcomm);
    MPI_Comm_rank(*groupcomm, rank);
    MPI_Barrier(*groupcomm);
}

void finishMPI(MPI_Comm* groupcomm)
{
    MPI_Barrier(*groupcomm);
    MPI_Comm_free(groupcomm);
    MPI_Finalize();
}

/*
 * execution
 * mpirun -np 4 ./orange -d full_path_name_of_xjd.xml 
 * 
 * this component expose...
 *  name: orangecomp
 *  dataport: orangeR1
 */
int main(int argc, char **argv)
{
    IC_XJD* xjd;
    
    float local_data[5];
    IC_Desc *desc;
    IC_Region* region1[2];

    MPI_Comm groupcomm;
    int rank, status;
    char* xjdname;

    /* initialize mpi and data*/
    fprintf(stderr, "initialize mpi and getting rank\n");
    initMPI(&argc, &argv, &rank, &groupcomm);

    fprintf(stderr, "initialize InterComm\n");    

    /* XJD name assignment should be done after MPI_Init*/
    if(parse_xjdname(argc, argv, &xjdname) < 0){
        fprintf(stderr, "XJD Option is required...\n");
        return -1;
    }

    if ( (xjd = IC_Initialize(COMPONENT_NAME, rank, xjdname, &status)) == NULL ){
        fprintf(stderr, "IC_Initialize failed at rank : %d\n", rank);        
        return -1;
    }

    /* Initialize local block and set global view*/
    init_data(rank, local_data);
    desc = create_block();

    /* create and register region*/
    create_region1(region1);
    IC_Register_region(xjd, region1, 2, REGION_NAME1, desc, local_data, &status);

    fprintf (stderr, "Before committing regions\n");
   
    /* all regions are registered. Now.. we can export or import    */
    IC_Commit_region(xjd, &status);

    fprintf (stderr, "Before Importing the regions\n");

    /* export region.*/
    status = IC_Import(xjd, REGION_NAME1);

    /* print result array */
    print_array(rank, local_data);

    /* free allocated memory*/
    fprintf(stderr, "clean up\n");
    IC_Free_region(region1[0]);
    IC_Free_region(region1[1]);
    IC_Free_desc(desc); 

    IC_Finalize(xjd, &status);

    fprintf(stderr, "finalize mpi\n");
    finishMPI(&groupcomm);
    fprintf(stderr, "No error with rank %d\n", rank);

    return 0;
}
